
To develop the Antlr-generated parser, download the Antlr tool (e.g., antlr-4.7.2-complete.jar) and add it to your classpath. See the [Installation](https://github.com/antlr/antlr4/blob/master/doc/getting-started.md#installation) section of the Antlr [Getting Started](https://github.com/antlr/antlr4/blob/master/doc/getting-started.md) page.

## To have Antlr generate a lexer and parser (in Java):

	antlr Cygnus.g4
	javac Cygnus*.java

(or `make`)

## Test using:

	grun Cygnus game tests/syntax.lp -gui

(or `make javatest`)

The -gui option visualizes the parse tree as an image.

or use the -tree option to print out the parse tree in LISP-like text form (root children):

	grun Cygnus game tests/syntax.lp -tree

## When you're satisfied with your Java lexer/parser, generate a Javascript version: 

	antlr -Dlanguage=JavaScript Cygnus.g4

(or `make js`)

This will generate the files: 

- CygnusParser.js - Builds the parse tree. Contains the parser class definition specific to grammar Cygnus that recognizes our Cygnus game language syntax. It contains a method for each rule in the grammar as well as some support code.
- CygnusLexer.js - contains the lexer class definition, which ANTLR generated by analyzing the lexical rules INT and WS as well as the grammar literals '{', ',', and '}'. Recall that the lexer tokenizes the input, breaking it up into vocabulary symbols. 
- .tokens files - ANTLR assigns a token type number to each token we define and stores these values in this file. It’s needed when we split a large grammar into multiple smaller grammars so that ANTLR can synchronize all the token type numbers.
- CygnusListener.js - By default, ANTLR parsers build a tree from the input. By walking that tree, a tree walker can fire “events” (callbacks) to a listener object that we provide. CygnusListener is the interface that describes the callbacks we can implement. This class makes it easy for us to override just the callbacks we’re interested in.
- .interp files - ?

Use `make clean` to remove unnecessary generated files.

